from typing import Optional

import pygame as pg

from ..base_element import UIElement
from viztools.utils import RenderContext, Color, Align


class Label(UIElement):
    def __init__(
            self, rect: pg.Rect, text: str, text_color: Color = (200, 200, 200),
            bg_color: Optional[Color] = None, align: Align = Align.CENTER
    ):
        super().__init__(rect)
        self._text = text
        self.text_color = text_color
        self.bg_color = bg_color
        self._text_surface: Optional[pg.Surface] = None
        self.align: Align = align

    def handle_event(self, event: pg.event.Event, render_context: RenderContext):
        super().handle_event(event, render_context)

    def set_text(self, text: str):
        self._text = text
        self._text_surface = None

    def render(self, screen: pg.Surface, render_context: RenderContext):
        if self.bg_color:
            pg.draw.rect(screen, self.bg_color, self.rect)

        if self._text:
            if self._text_surface is None:
                self._text_surface = render_context.font.render(self._text, True, self.text_color)
            rect = self.align.arrange_in_rect(self._text_surface.get_rect(), self.rect)
            screen.blit(self._text_surface, rect)

    def update(self, render_context: RenderContext):
        pass



import io
from pathlib import Path
from typing import Union, Optional, Tuple

import numpy as np
import pygame as pg
from PIL import Image as PilImage
from PIL.Image import Transpose

from viztools.coordinate_system import CoordinateSystem
from viztools.drawable.base_drawable import Drawable
from viztools.utils import RenderContext, Align


class Image(Drawable):
    def __init__(
            self, image: Union[np.ndarray, PilImage.Image, str, Path], position: np.ndarray,
            size: Union[np.ndarray, float] = 0.01, align: Align = Align.CENTER,
            offset: Optional[np.ndarray] = None, offset_color: Optional[np.ndarray] = None
    ):
        """
        Initializes a list of lines.

        :param image: Numpy array of shape [h, w, 3] or pillow image.
        :param position: The position of the image as a numpy array of shape [2].
        :param size: The size of the image as a numpy array or tuple (height, width) in world coordinates,
            or it can be defined as a scale factor to original size.
        :param align: The type of anker to use.
        :param offset: The offset of the image as a numpy array of shape [2].
        :param offset_color: The color of the offset as a numpy array of shape [3] or [4] (with alpha).
        """
        super().__init__()

        image = to_pil_image(image)
        image = fix_image_axis_swap(image)
        self.image_data: bytes = encode_image(image)
        self.image_surface: Optional[pg.Surface] = None
        self.position = position
        self.align = align
        if isinstance(size, float):
            size = np.array(get_image_size(image)) * size
        elif isinstance(size, (tuple, np.ndarray)):
            size = np.asarray(size)
        else:
            raise TypeError(f'size must be a float, tuple or numpy array, not {type(size)}.')
        self.size = size
        self.last_size = None
        self.scaled_surface = None
        self.offset = offset
        self.offset_color = offset_color

    def _ensure_image_surface(self):
        if self.image_surface is None:
            self.image_surface = bytes_to_surf_array(self.image_data)

    def render(self, screen: pg.Surface, coordinate_system: CoordinateSystem, render_context: RenderContext):
        anchor_point = coordinate_system.space_to_screen_t(self.position).flatten().astype(int)
        pos = self.position + self.offset if self.offset is not None else self.position
        screen_points = coordinate_system.space_to_screen_t(pos).flatten().astype(int)
        size = np.abs(coordinate_system.space_to_screen_t(self.size, translate=False).flatten().astype(int))

        # get target rect
        target_rect = pg.Rect(0, 0, size[1], size[0])
        target_rect = self.align.arrange_by_anker(target_rect, screen_points)

        if target_rect.colliderect(screen.get_rect()) and np.prod(target_rect.size) < 30000000:
            self._ensure_image_surface()
            if self.last_size is None or np.any(size != self.last_size):
                self.scaled_surface = pg.transform.scale(self.image_surface, (size[1], size[0]))
                self.last_size = size

            if self.offset_color is not None:
                pg.draw.line(screen, self.offset_color, anchor_point, screen_points, 2)
            screen.blit(self.scaled_surface, target_rect)
        else:
            # remove unused data
            self.last_size = None
            self.scaled_surface = None
            self.image_surface = None

    def handle_event(
            self, event: pg.event.Event, screen: pg.Surface, coordinate_system: CoordinateSystem,
            render_context: RenderContext
    ):
        pass

    def update(
            self, screen: pg.Surface, coordinate_system: CoordinateSystem, render_context: RenderContext
    ):
        pass


def fix_image_axis_swap(image: Union[np.ndarray, PilImage.Image]) -> Union[np.ndarray, PilImage.Image]:
    """
    Swapes axis for the given image.
    :param image: The image to swap.
    :return: The image with swapped axes.
    """
    if isinstance(image, np.ndarray):
        return image.swapaxes(0, 1)
    elif isinstance(image, PilImage.Image):
        return image.transpose(Transpose.TRANSPOSE)
    else:
        raise TypeError(f'image must be of type np.ndarray or PIL.Image, not {type(image)}.')


def get_image_size(image: PilImage.Image) -> Tuple[int, int]:
    """
    Returns the size of the image in pixels as [height, width].

    :param image: The image to get the size of.
    :return: The size of the image in pixels as [height, width].
    """
    if isinstance(image, PilImage.Image):
        return image.size
    else:
        raise TypeError(f'image must be of type np.ndarray or PIL.Image, not {type(image)}.')


def to_pil_image(image: Union[np.ndarray, PilImage.Image, str, Path]) -> PilImage.Image:
    """
    Convert to pillow image. Accepts numpy array, path to image or pillow image.

    :param image: The image to convert.
    :return: A pillow image.
    """
    if isinstance(image, (str, Path)):
        return PilImage.open(image)
    elif isinstance(image, PilImage.Image):
        return image
    elif isinstance(image, np.ndarray):
        return PilImage.fromarray(image)
    else:
        raise TypeError(f'image must be of type str, Path or np.ndarray, not {type(image)}.')


def encode_image(image: PilImage.Image) -> bytes:
    """
    Encode data into bytes (jpg) for small memory footprint.

    :param image: The image to encode.
    :return: The jpg encoded image as bytes.
    """
    with io.BytesIO() as bytes_stream:
        image.save(bytes_stream, format="JPEG")
        return bytes_stream.getvalue()


def bytes_to_surf_array(data: bytes) -> pg.Surface:
    return pil_to_surf_array(decode_image(data))


def decode_image(data: bytes) -> PilImage.Image:
    return PilImage.open(io.BytesIO(data))


def pil_to_surf_array(pil_img: PilImage.Image) -> pg.Surface:
    """
    Converts a PIL Image to a Pygame Surface using surf_array.
    """
    pil_img = pil_img.convert('RGB')
    image_array = np.asarray(pil_img, dtype=np.uint8)
    pygame_surface = pg.surfarray.make_surface(image_array)
    return pygame_surface.convert()



from abc import ABC
from typing import Tuple, Optional, List, Union, Iterable

import numpy as np
import pygame as pg

from viztools.controller.coordinate_system_controller import CoordinateSystemController
from viztools.coordinate_system import CoordinateSystem, draw_coordinate_system
from viztools.drawable import Drawable
from viztools.ui.container.base_container import Container
from viztools.ui.elements.base_element import UIElement
from viztools.utils import RenderContext, DEFAULT_FONT_SIZE


class Viewer(ABC):
    def __init__(
            self, screen_size: Optional[Tuple[int, int]] = None, title: str = "Visualization", framerate: int = 60,
            font_size: int = DEFAULT_FONT_SIZE, drag_mouse_button: int = 2
    ):
        pg.init()
        pg.scrap.init()
        pg.key.set_repeat(130, 25)
        mode = pg.RESIZABLE
        if screen_size is None:
            screen_size = (0, 0)
        if screen_size == (0, 0):
            mode = mode | pg.FULLSCREEN
        self.screen = pg.display.set_mode(screen_size, mode)
        pg.display.set_caption(title)

        self.running = True
        self.clock = pg.time.Clock()
        self.framerate = framerate
        self.mouse_pos = np.array(pg.mouse.get_pos(), dtype=np.int32)

        self.coordinate_system = CoordinateSystem(screen_size)
        self.coordinate_system_controller = CoordinateSystemController(
            self.coordinate_system, drag_mouse_button=drag_mouse_button
        )

        self.render_context = RenderContext.default(font_size)

        self._drawable_cache: Optional[List[Drawable]] = None
        self._ui_element_cache: Optional[List[Union[Container, UIElement]]] = None

    def iter_ui_elements(self) -> Iterable[Union[UIElement, Container]]:
        """
        Iter over all elements in the container.
        :return: Iterable of BaseElement objects.
        """
        if self._ui_element_cache is None:
            self._ui_element_cache = [
                elem for elem in self.__dict__.values() if isinstance(elem, (UIElement, Container))
            ]

        yield from self._ui_element_cache

    def iter_drawables(self) -> Iterable[Drawable]:
        """
        Iter over all elements in the container.
        :return: Iterable of BaseElement objects.
        """
        if self._drawable_cache is None:
            self._drawable_cache = [elem for elem in self.__dict__.values() if isinstance(elem, Drawable)]

        yield from self._drawable_cache

    def run(self):
        while self.running:
            self.handle_events()
            self.update()
            self.render()
            self.clock.tick(self.framerate)
        pg.quit()

    def render_content(self):
        self.render_drawables(self.iter_drawables())

    def render_ui(self):
        self.render_ui_elements(self.iter_ui_elements())

    def render_drawables(self, drawables: Iterable[Drawable]):
        for drawable in drawables:
            drawable.draw(self.screen, self.coordinate_system, self.render_context)

    def render_ui_elements(self, ui_elements: Iterable[Union[UIElement, Container]]):
        for ui_element in ui_elements:
            ui_element.draw(self.screen, self.render_context)

    def render_coordinate_system(self, draw_numbers=True):
        draw_coordinate_system(self.screen, self.coordinate_system, self.render_context.font, draw_numbers=draw_numbers)

    def render(self):
        self.render_coordinate_system(draw_numbers=True)
        self.render_content()
        self.render_ui()
        pg.display.flip()

    def handle_events(self):
        events = pg.event.get()
        for event in events:
            self.handle_event(event)
        for ui_element in self.iter_ui_elements():
            ui_element.handle_events(events, self.render_context)
        for drawable in self.iter_drawables():
            drawable.handle_events(events, self.screen, self.coordinate_system, self.render_context)

    def handle_event(self, event: pg.event.Event):
        self.coordinate_system_controller.handle_event(event)
        if event.type == pg.MOUSEMOTION:
            self.mouse_pos = np.array(event.pos)
        if event.type == pg.QUIT:
            self.running = False

    def update(self):
        pass



from typing import Iterable, Optional, List

import pygame as pg

from ..elements.base_element import UIElement
from viztools.utils import RenderContext


class Container:
    def __init__(self, visible: bool = True):
        self.visible = visible
        self._element_cache: Optional[List[UIElement]] = None

    def iter_elements(self) -> Iterable[UIElement]:
        """
        Iter over all elements in the container.
        :return: Iterable of BaseElement objects.
        """
        if self._element_cache is None:
            self._element_cache = [elem for elem in self.__dict__.values() if isinstance(elem, UIElement)]

        yield from self._element_cache

    def handle_events(self, events: List[pg.event.Event], render_context: RenderContext):
        if self.visible:
            for elem in self.iter_elements():
                elem.handle_events(events, render_context)

    def draw(self, screen: pg.Surface, render_context: RenderContext):
        if self.visible:
            for element in self.iter_elements():
                element.draw(screen, render_context)



from itertools import pairwise
from typing import Tuple

import numpy as np
import pygame as pg

from viztools.coordinate_system import CoordinateSystem
from viztools.drawable.base_drawable import Drawable
from viztools.utils import RenderContext


class Lines(Drawable):
    def __init__(self, points: np.ndarray, color: np.ndarray = None):
        """
        Initializes a list of lines.

        :param points: Numpy array of shape [N, 2] where N is the number of points.
        :param color: The color of the lines as numpy array of shape [3] or [4] (with alpha).
        """
        super().__init__()
        self.points = points
        self.color = color

    def render(self, screen: pg.Surface, coordinate_system: CoordinateSystem, render_context: RenderContext):
        screen_points = coordinate_system.space_to_screen_t(self.points)
        for p1, p2 in pairwise(screen_points):
            pg.draw.line(screen, self.color, p1, p2)

    def clicked_points(
            self, event: pg.event.Event, coordinate_system: CoordinateSystem, max_distance: float = 10.0
    ) -> np.ndarray:
        """
        Returns the indices of the points clicked by the mouse. Returns an empty array if no point was clicked.

        :param event: The event to check.
        :param coordinate_system: The coordinate system to use.
        """
        if event.type == pg.MOUSEBUTTONDOWN and event.button == 1:
            return self.hovered_points(np.array(event.pos, dtype=np.int32), coordinate_system, max_distance)
        return np.array([])

    def hovered_points(
            self, mouse_pos: np.ndarray, coordinate_system: CoordinateSystem, max_distance: float = 10.0
    ) -> np.ndarray:
        screen_pos = mouse_pos.reshape(1, 2)
        screen_points = coordinate_system.space_to_screen_t(self.points)
        distances = np.linalg.norm(screen_points - screen_pos, axis=1)
        return np.nonzero(distances < max_distance)[0]

    def closest_point(self, pos: np.ndarray, coordinate_system: CoordinateSystem) -> Tuple[int, float]:
        """
        Finds the closest point to the given position.

        This function calculates the closest point to a specified 2D position on
        the screen.

        :param pos: The 2D position in screen coordinates to calculate the distance from.
        :param coordinate_system: The coordinate system used for transforming space
                                  coordinates to screen coordinates.
        :return: A tuple containing the index of the closest point and the distance
                 to that closest point.
        :rtype: Tuple[int, float]
        """
        screen_pos = pos.reshape(1, 2)
        screen_points = coordinate_system.space_to_screen_t(self.points)
        distances = np.linalg.norm(screen_points - screen_pos, axis=1)
        closest_index = np.argmin(distances)
        return int(closest_index), max(float(distances[closest_index]), 0.0)

    def handle_event(
            self, event: pg.event.Event, screen: pg.Surface, coordinate_system: CoordinateSystem,
            render_context: RenderContext
    ):
        pass

    def update(self, screen: pg.Surface, coordinate_system: CoordinateSystem, render_context: RenderContext):
        pass



from .base_drawable import Drawable
from .implementations.image import Image
from .implementations.lines import Lines
from .implementations.overlay_text import OverlayText
from .implementations.points import Points

__all__ = ['Drawable', 'Image', 'Lines', 'OverlayText', 'Points']



from abc import ABC, abstractmethod
from typing import List, final

import pygame as pg

from viztools.coordinate_system import CoordinateSystem
from viztools.utils import RenderContext


class Drawable(ABC):
    def __init__(self):
        self.render_needed = True

    @final
    def handle_events(
            self, events: List[pg.event.Event], screen: pg.Surface, coordinate_system: CoordinateSystem,
            render_context: RenderContext
    ):
        """
        Handles the given events and updates the element, if needed.
        If a redraw is necessary, sets self.render_needed to True.

        :param events: The events to handle.
        :param screen: The screen that will be used for drawing.
        :param coordinate_system: The coordinate system, this drawable is rendered in.
        """
        for event in events:
            self.handle_event(event, screen, coordinate_system, render_context)
        self.update(screen, coordinate_system, render_context)

    @abstractmethod
    def handle_event(
            self, event: pg.event.Event, screen: pg.Surface, coordinate_system: CoordinateSystem,
            render_context: RenderContext
    ):
        """
        Handles the given events and updates the element.
        :param event: The event to handle.
        :param screen: The screen that will be used for drawing.
        :param coordinate_system: The coordinate system, this drawable is rendered in.
        :param render_context: The render context used for drawing.
        """
        pass

    @abstractmethod
    def update(self, screen: pg.Surface, coordinate_system: CoordinateSystem, render_context: RenderContext):
        """
        Updates the element. Also sets self.render_needed to True, if needed.

        :param screen: The screen that will be used for drawing.
        :param coordinate_system: The coordinate system, this drawable is rendered in.
        """
        pass

    @final
    def draw(self, screen: pg.Surface, coordinate_system: CoordinateSystem, render_context: RenderContext):
        """
        Draws the element to the screen.

        :param screen: The screen that will be used for drawing.
        :param coordinate_system: The coordinate system, this drawable is rendered in.
        :param render_context: The render context used for drawing.
        """
        self.render(screen, coordinate_system, render_context)
        self.finalize()

    @abstractmethod
    def render(self, screen: pg.Surface, coordinate_system: CoordinateSystem, render_context: RenderContext):
        """
        Draws the element to the screen.

        :param screen: The screen that will be used for drawing.
        :param coordinate_system: The coordinate system, this drawable is rendered in.
        :param render_context: The render context used for drawing.
        """
        pass

    def finalize(self):
        """
        Finalize this drawable.
        """
        pass



import enum
import warnings
from typing import Union, Tuple

import pygame as pg
import numpy as np

DEFAULT_FONT_SIZE = 16


def to_np_array(p):
    if isinstance(p, np.ndarray):
        return p
    return np.array(p)


Color = Union[np.ndarray, Tuple[int, int, int, int], Tuple[int, int, int]]


def normalize_color(color: Color) -> np.ndarray:
    if len(color) == 3:
        return np.array([*color, 255], dtype=np.float32)
    if len(color) != 4:
        raise ValueError(f'color must be of length 3 or 4, not {len(color)}.')
    return np.array(color, dtype=np.float32)


class RenderContext:
    def __init__(self, font: pg.font.Font):
        self.font = font
        self.mouse_pressed = False

    @staticmethod
    def default(font_size: int = DEFAULT_FONT_SIZE):
        font = load_font(font_size)
        return RenderContext(font)


class Align(enum.StrEnum):
    CENTER = 'center'
    LEFT = 'left'
    RIGHT = 'right'
    TOP = 'top'
    BOTTOM = 'bottom'
    TOP_LEFT = 'top_left'
    TOP_RIGHT = 'top_right'
    BOTTOM_LEFT = 'bottom_left'
    BOTTOM_RIGHT = 'bottom_right'

    def set_rect(self, rect: pg.Rect, position: Tuple[int, int]):
        if self == Align.CENTER:
            rect.center = position
        elif self == Align.LEFT:
            rect.midleft = position
        elif self == Align.RIGHT:
            rect.midright = position
        elif self == Align.TOP:
            rect.midtop = position
        elif self == Align.BOTTOM:
            rect.midbottom = position
        elif self == Align.TOP_LEFT:
            rect.topleft = position
        elif self == Align.TOP_RIGHT:
            rect.topright = position
        elif self == Align.BOTTOM_LEFT:
            rect.bottomleft = position
        elif self == Align.BOTTOM_RIGHT:
            rect.bottomright = position
        else:
            raise ValueError(f'unknown anker type: {self}')

    def get_pos(self, rect: pg.Rect) -> Tuple[int, int]:
        if self == Align.CENTER:
            return rect.center
        elif self == Align.LEFT:
            return rect.midleft
        elif self == Align.RIGHT:
            return rect.midright
        elif self == Align.TOP:
            return rect.midtop
        elif self == Align.BOTTOM:
            return rect.midbottom
        elif self == Align.TOP_LEFT:
            return rect.topleft
        elif self == Align.TOP_RIGHT:
            return rect.topright
        elif self == Align.BOTTOM_LEFT:
            return rect.bottomleft
        elif self == Align.BOTTOM_RIGHT:
            return rect.bottomright
        else:
            raise ValueError(f'unknown anker type: {self}')

    def arrange_by_anker(self, rect: pg.Rect, anker: Union[np.ndarray, Tuple[int, int]]) -> pg.Rect:
        new_rect = rect.copy()
        self.set_rect(new_rect, anker)
        return new_rect

    def arrange_in_rect(self, rect: pg.Rect, container: pg.Rect) -> pg.Rect:
        new_rect = rect.copy()
        self.set_rect(new_rect, self.get_pos(container))
        return new_rect


def load_font(font_size: int = DEFAULT_FONT_SIZE):
    """
    Helper function to load the default font.
    :return: The font to use
    """
    try:
        font = pg.font.Font(None, font_size)
    except pg.error:
        warnings.warn("Warning: Could not load default font. Text will not be rendered.")
        font = None
    return font


def clamp(n, minn, maxn):
    return max(min(maxn, n), minn)



from typing import Tuple
import numbers

import pygame as pg
import numpy as np

from viztools.utils import to_np_array


TARGET_NUM_POINTS = 12
TARGET_DIVIDENDS = [1, 2.5, 5, 10]
DEFAULT_SCREEN_SIZE = (1280, 720)


class CoordinateSystem:
    def __init__(self, screen_size: Tuple[int, int] | np.ndarray):
        screen_size = to_np_array(screen_size)
        coord = create_affine_transformation(screen_size/2, (100, -100))
        self.coord: np.ndarray = coord
        self.inverse_coord: np.ndarray = np.linalg.pinv(self.coord)

        # user control
        self.zoom_factor = 100

    def zoom_out(self, focus_point=None, scale=1.2):
        scale = 1 / scale
        self.zoom_factor *= scale
        scale_mat = create_affine_transformation(scale=scale)
        self.coord = self.coord @ scale_mat
        if focus_point is not None:
            translation = (focus_point - self.get_zero_screen_point().flatten()) * (1 - scale)
            self.translate(translation)
        self.update_inv()

    def zoom_in(self, focus_point=None, scale=1.2):
        self.zoom_factor *= scale
        scale_mat = create_affine_transformation(scale=scale)
        self.coord = self.coord @ scale_mat
        if focus_point is not None:
            translation = (focus_point - self.get_zero_screen_point().flatten()) * (1 - scale)
            self.translate(translation)
        self.update_inv()

    def translate(self, direction):
        direction *= np.array([1, -1])
        translation_mat = create_affine_transformation(translation=direction / self.coord[0, 0])
        self.coord = self.coord @ translation_mat
        self.update_inv()

    def get_zero_screen_point(self):
        """
        Get the zero point of the coordinate system in screen coordinates.
        """
        return self.space_to_screen(np.array([0.0, 0.0]))

    def get_viewport(self, screen_size: Tuple[int, int] | np.ndarray) -> np.ndarray:
        """
        Calculates the viewport for the given screen size in world coordinates.

        This method transforms the screen size dimensions into the space coordinates.
        It is capable of handling both tuples and NumPy arrays as input for the screen size dimensions.

        :param screen_size: The dimensions of the screen size, given as either a tuple of
            integers (width, height) or a NumPy array.
            Type: Tuple[int, int] | np.ndarray
        :return: Transformed viewport represented as a NumPy array after applying
            screen-to-space transformation.
            Type: np.ndarray
        """
        return self.screen_to_space_t(np.array([[0.0, 0.0], screen_size]))

    def space_to_screen_t(self, mat: np.ndarray, translate: bool = True) -> np.ndarray:
        """
        Transform the given matrix with the internal coordinates.

        :param mat: A list of column vectors with shape [N, 2]. For vectors shape should be [1, 2].
        :param translate: If translate is True, translation is applied. Otherwise only rotation and scaling is applied.
        :return: A list of column vectors with shape [N, 2].
        """
        return self.space_to_screen(mat.T, translate).T

    def space_to_screen(self, mat: np.ndarray, translate: bool = True) -> np.ndarray:
        """
        Transform the given matrix with the internal coordinates.

        :param mat: A list of column vectors with shape [2, N]. For vectors shape should be [2, 1].
        :param translate: If translate is True, translation is applied. Otherwise only rotation and scaling is applied.
        :return: A list of column vectors with shape [2, N].
        """
        mat = to_np_array(mat)
        if mat.shape == (2,):
            mat = mat.reshape(2, 1)
        return transform(self.coord, mat, translate=translate)

    def screen_to_space_t(self, mat: np.ndarray, translate: bool = True) -> np.ndarray:
        """
        Transform screen coordinates to space coordinates using the inverse
        coordinate transformation matrix.

        :param mat: Input matrix representing screen coordinates. It can be
            either a 2D array with shape (N, 2) or a 1D array with shape (2,).
            If the shape is (2,), it will be reshaped to (1, 2).
        :param translate: If translate is True, translation is applied. Otherwise only rotation and scaling is applied.

        :return: A transformed matrix in space coordinates based on the
            inverse coordinate transformation matrix of the object.
        """
        return self.screen_to_space(mat.T, translate).T

    def screen_to_space(self, mat: np.ndarray, translate: bool = True):
        """
        Transform screen coordinates to space coordinates using the inverse
        coordinate transformation matrix.

        :param mat: Input matrix representing screen coordinates. It can be
            either a 2D array with shape (2, N) or a 1D array with shape (2,).
            If the shape is (2,), it will be reshaped to (2, 1).
        :param translate: If translate is True, translation is applied. Otherwise only rotation and scaling is applied.

        :return: A transformed matrix in space coordinates based on the
            inverse coordinate transformation matrix of the object.
        """
        mat = to_np_array(mat)
        if mat.shape == (2,):
            mat = mat.reshape(2, 1)
        return transform(self.inverse_coord, mat, translate=translate)

    def update_inv(self):
        self.inverse_coord = np.linalg.pinv(self.coord)


def transform(
        transform_matrix: np.ndarray, mat: np.ndarray, perspective: bool = False, translate: bool = True
) -> np.ndarray:
    """
    Transforms a given matrix with the given transformation matrix.
    Transformation matrix should be of shape [2, 2] or [3, 3]. If transformation matrix is of shape [3, 3] and the
    matrix to transform is of shape [2, N], matrix will be padded with ones to shape [3, N].
    If mat is of shape [2,] it will be converted to [2, 1].

    The calculation will be transform_matrix @ mat.

    :param transform_matrix: A np.ndarray with shape [2, 2] or [3, 3].
    :param mat: The matrix to convert of shape [2, N]. If mat is of shape [2,] it will be converted to [2, 1].
    :param perspective: If perspective is True and the transform_mat is of shape (3, 3), the x- and y-axis of the
                        resulting vector are divided by the resulting z axis.
    :param translate: If translate is True, the resulting vector is translated by the translation vector of the
                      transform_matrix. Otherwise only rotation and scaling is applied.
    :return:
    """
    expanded = False
    if mat.shape == (2,):
        mat = mat.reshape((2, 1))
        expanded = True

    padded = False
    if transform_matrix.shape == (3, 3):
        # noinspection PyTypeChecker
        pad_func = np.ones if translate else np.zeros
        mat = np.concatenate([mat, pad_func((1, mat.shape[1]))], axis=0)
        padded = True

    result = transform_matrix @ mat

    if expanded:
        result = result[:, 0]

    if padded:
        if perspective:
            result = result[:-1] / result[-1]
        else:
            result = result[:-1]
    return result


def create_affine_transformation(
        translation: float | np.ndarray | Tuple[int, int] = 0, scale: float | np.ndarray | Tuple[float, float] = 1
) -> np.ndarray:
    if isinstance(scale, numbers.Number):
        scale = (scale, scale)
    scale_coord = np.array(
        [[scale[0], 0, 0],
         [0, scale[1], 0],
         [0, 0, 1]]
    )
    if isinstance(translation, numbers.Number):
        translation = (translation, translation)
    translate_coord = np.array(
        [[1, 0, translation[0]],
         [0, 1, translation[1]],
         [0, 0, 1]]
    )
    return translate_coord @ scale_coord


def draw_coordinate_system(
        screen: pg.Surface, coordinate_system: CoordinateSystem, render_font: pg.font.Font, draw_numbers=True
):
    screen.fill((0, 0, 0))

    def adapt_quotient(quotient):
        if quotient <= 0:
            raise ValueError('Invalid quotient: {}'.format(quotient))
        numb_ten_potency = 0
        while quotient > 10:
            quotient *= 0.1
            numb_ten_potency += 1
        while quotient < 1:
            quotient *= 10
            numb_ten_potency -= 1

        diffs = [abs(quotient - target) for target in TARGET_DIVIDENDS]
        index = np.argmin(diffs)
        best_fitting = TARGET_DIVIDENDS[index] * (10 ** numb_ten_potency)

        return best_fitting

    width, height = screen.get_size()
    extreme_points = np.array([
        [0, 0],
        [width, height]
    ]).T
    extreme_points = coordinate_system.screen_to_space(extreme_points).T
    target_num_points = TARGET_NUM_POINTS * width // DEFAULT_SCREEN_SIZE[0]
    target_dividend = (extreme_points[1, 0] - extreme_points[0, 0]) / target_num_points
    dividend = adapt_quotient(target_dividend)
    x_minimum = np.round(extreme_points[0, 0] / dividend) * dividend
    x_maximum = np.round(extreme_points[1, 0] / dividend) * dividend
    x_points = np.arange(x_minimum, x_maximum + dividend, dividend)
    for x in x_points:
        vertical_lines = np.array([[x, 0], [x, 0]])
        transformed_vertical_lines = coordinate_system.space_to_screen(vertical_lines.T).T
        transformed_vertical_lines[:, 1] = [0, height]
        color = np.array([30, 30, 30])
        if x == 0:
            color = np.array([50, 50, 50])
        pg.draw.line(screen, color, transformed_vertical_lines[0], transformed_vertical_lines[1])

    y_minimum = np.round(extreme_points[1, 1] / dividend) * dividend
    y_maximum = np.round(extreme_points[0, 1] / dividend) * dividend
    y_points = np.arange(y_minimum, y_maximum + dividend, dividend)

    for y in y_points:
        horizontal_lines = np.array([[extreme_points[0, 0], y], [extreme_points[1, 0], y]])
        transformed_horizontal_lines = coordinate_system.space_to_screen(horizontal_lines.T).T
        transformed_horizontal_lines[:, 0] = [0, width]
        color = np.array([30, 30, 30])
        if y == 0:
            color = np.array([50, 50, 50])
        pg.draw.line(screen, color, transformed_horizontal_lines[0], transformed_horizontal_lines[1])

    # draw numbers
    if draw_numbers:
        zero_point = coordinate_system.space_to_screen(np.array([0, 0]))

        if 0 < zero_point[1] < height:
            for x in x_points:
                if abs(x) > 10 ** -5:
                    float_format = '{:.2f}' if abs(x) > 1 else '{:.2}'
                    font = render_font.render(
                        float_format.format(x), True, np.array([120, 120, 120]), np.array([0, 0, 0, 0])
                    )
                    pos = coordinate_system.space_to_screen(np.array([x, 0]))
                    pos += 10
                    # noinspection PyTypeChecker
                    render_pos: Tuple[int, int] = tuple(pos.flatten().tolist())
                    screen.blit(font, render_pos)

        if 0 < zero_point[0] < width:
            for y in y_points:
                if abs(y) > 10 ** -5:
                    float_format = '{:.2f}' if abs(y) > 1 else '{:.2}'
                    font = render_font.render(
                        float_format.format(y), True, np.array([120, 120, 120]), np.array([0, 0, 0, 0])
                    )
                    pos = coordinate_system.space_to_screen(np.array([0, y]))
                    pos += 10
                    # noinspection PyTypeChecker
                    render_pos: Tuple[int, int] = tuple(pos.flatten().tolist())
                    screen.blit(font, render_pos)



from viztools.ui.elements.implementations.label import Label
from viztools.ui.elements.implementations.button import Button
from viztools.ui.elements.implementations.edit_field import EditField
from viztools.ui.elements.implementations.text_field import TextField

__all__ = ['Label', 'Button', 'EditField', 'TextField']
